from typing import Any


def encoding_to_types(encoding: list[dict[str, Any]]) -> list[str]:
    def parse_type(item: dict[str, Any]) -> str:
        if item["type"].startswith("tuple"):
            inner_types = ",".join(parse_type(comp) for comp in item["components"])
            return f"({inner_types}){item['type'][5:]}"
        return item["type"]

    return [parse_type(item) for item in encoding]



from ..releases import RELEASES
from ..assets import define_asset_list
from ..networks import NETWORK_BY_SLUG
from ..price_feeds import RateAsset


SULU = RELEASES["arbitrum"]["sulu"]


ASSETS = define_asset_list(
    NETWORK_BY_SLUG["arbitrum"],
    [
        {
            "decimals": 18,
            "id": "0x040d1EdC9569d4Bab2D15287Dc5A4F10F56a56B8",
            "name": "Balancer",
            "releases": [SULU],
            "symbol": "BAL",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xBE5eA816870D11239c543F84b71439511D70B94f",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x354A6dA3fcde098F8389cad84b0182725c6C91dE",
            "name": "Compound",
            "releases": [SULU],
            "symbol": "COMP",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xe7C53FFd03Eb6ceF7d208bC4C13446c76d1E5884",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x11cDb42B0EB46D95f990BeDD4695A6e3fA034978",
            "name": "Curve DAO Token",
            "releases": [SULU],
            "symbol": "CRV",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xaebDA2c976cfd1eE1977Eac079B4382acb849325",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xaAFcFD42c9954C6689ef1901e03db742520829c5",
            "name": "Convex Token",
            "releases": [SULU],
            "symbol": "CVX",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x851175a919f36c8e30197c09a9A49dA932c2CC00",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x9623063377AD1B27544C965cCd7342f7EA7e88C7",
            "name": "Graph Token",
            "releases": [SULU],
            "symbol": "GRT",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x0F38D86FceF4955B705F35c9e41d1A16e0637c73",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x1DEBd73E752bEaF79865Fd6446b0c970EaE7732f",
            "name": "Coinbase Wrapped Staked ETH",
            "releases": [SULU],
            "symbol": "cbETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xa668682974E3f121185a3cD94f00322beC674275",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x2416092f143378750bb29b79eD961ab195CcEea5",
            "name": "Renzo Restaked ETH",
            "releases": [SULU],
            "symbol": "ezETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x11E1836bFF2ce9d6A5bec9cA79dc998210f3886d",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 9,
            "id": "0x2bcC6D6CdBbDC0a4071e48bb3B969b06B3330c07",
            "name": "Wrapped SOL",
            "releases": [SULU],
            "symbol": "SOL",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x24ceA4b8ce57cdA5058b924B9B9987992450590c",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 8,
            "id": "0x2f2a2543B76A4166549F7aaB2e75Bef0aefC5B0f",
            "name": "Wrapped BTC",
            "releases": [SULU],
            "symbol": "WBTC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xd0C7101eACbB49F3deCcCc166d238410D6D46d57",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x3082CC23568eA640225c2467653dB90e9250AaA0",
            "name": "Radiant",
            "releases": [SULU],
            "symbol": "RDNT",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x20d0Fcab0ECFD078B036b6CAf1FaC69A6453b352",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x35751007a407ca6FEFfE80b3cB397736D2cf4dbe",
            "name": "Wrapped eETH",
            "releases": [SULU],
            "symbol": "weETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xE141425bc1594b8039De6390db1cDaf4397EA22b",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x4186BFC76E2E237523CBC30FD220FE055156b41F",
            "name": "KelpDao Restaked ETH",
            "releases": [SULU],
            "symbol": "rsETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb0EA543f9F8d4B818550365d13F66Da747e1476A",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x565609fAF65B92F7be02468acF86f8979423e514",
            "name": "Wrapped AVAX",
            "releases": [SULU],
            "symbol": "WAVAX",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x8bf61728eeDCE2F32c456454d87B5d6eD6150208",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x5979D7b546E38E414F7E9822514be443A4800529",
            "name": "Wrapped liquid staked Ether 2.0",
            "releases": [SULU],
            "symbol": "wstETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb523AE262D20A936BC152e6023996e46FDC2A95D",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x82aF49447D8a07e3bd95BD0d56f35241523fBab1",
            "name": "Wrapped Ether",
            "releases": [SULU],
            "symbol": "WETH",
            "type": "primitive",
            "price_feed": {
                "type": "WETH",
            },
        },
        {
            "decimals": 18,
            "id": "0x8f5c1A99b1df736Ad685006Cb6ADCA7B7Ae4b514",
            "name": "Melon Token",
            "releases": [SULU],
            "symbol": "MLN",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xD07de6e37A011CCAfD375d7eb130205E0fa24d69",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x912CE59144191C1204E64559FE8253a0e49E6548",
            "name": "Arbitrum",
            "releases": [SULU],
            "symbol": "ARB",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb2A824043730FE05F3DA2efaFa1CBbe83fa548D6",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xa9004A5421372E1D83fB1f85b0fc986c912f91f3",
            "name": "Wrapped BNB",
            "releases": [SULU],
            "symbol": "WBNB",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x6970460aabF80C5BE983C6b74e5D06dEDCA95D4A",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xaC800FD6159c2a2CB8fC31EF74621eB430287a5A",
            "name": "Optimism",
            "releases": [SULU],
            "symbol": "OP",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x205aaD468a11fd5D34fA7211bC6Bad5b3deB9b98",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0xaf88d065e77c8cC2239327C5EDb3A432268e5831",
            "name": "USD Coin",
            "releases": [SULU],
            "symbol": "USDC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x50834F3163758fcC1Df9973b6e91f0F0F0434aD3",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xba5DdD1f9d7F570dc94a51479a000E3BCE967196",
            "name": "Aave Token",
            "releases": [SULU],
            "symbol": "AAVE",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xaD1d5344AaDE45F43E596773Bcc4c423EAbdD034",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xbc011A12Da28e8F0f528d9eE5E7039E22F91cf18",
            "name": "swETH",
            "releases": [SULU],
            "symbol": "swETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x05Bc6e5Fb110589bb366A3Cd7CdBe143EeBA2168",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1",
            "name": "Dai Stablecoin",
            "releases": [SULU],
            "symbol": "DAI",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xc5C8E77B397E531B8EC06BFb0048328B30E9eCfB",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xEC70Dcb4A1EFa46b8F2D97C310C9c4790ba5ffA8",
            "name": "Rocket Pool ETH",
            "releases": [SULU],
            "symbol": "rETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xD6aB2298946840262FcC278fF31516D39fF611eF",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xf97f4df75117a78c1A5a0DBb814Af92458539FB4",
            "name": "ChainLink Token",
            "releases": [SULU],
            "symbol": "LINK",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb7c8Fb1dB45007F98A68Da0588e1AA524C317f27",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xFa7F8980b0f1E64A2062791cc3b0871572f1F7f0",
            "name": "Uniswap",
            "releases": [SULU],
            "symbol": "UNI",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x9C917083fDb403ab5ADbEC26Ee294f6EcAda2720",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xfc5A1A6EB076a2C7aD06eD22C90d7E710E35ad0a",
            "name": "GMX",
            "releases": [SULU],
            "symbol": "GMX",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xDB98056FecFff59D032aB628337A4887110df3dB",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0xFd086bC7CD5C481DCC9C85ebE478A1C0b69FCbb9",
            "name": "USD₮0",
            "releases": [SULU],
            "symbol": "USD₮0",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x3f3f5dF88dC9F13eac63DF89EC16ef6e7E25DdE7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xe50fA9b3c56FfB159cB0FCA61F5c9D750e8128c8",
            "name": "Aave Arbitrum WETH",
            "releases": [SULU],
            "symbol": "aArbWETH",
            "type": "aave-v3",
            "underlying": "0x82aF49447D8a07e3bd95BD0d56f35241523fBab1",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x639Fe6ab55C921f74e7fac1ee960C0B6293ba612",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0x724dc807b04555b71ed48a6896b6F41593b8C637",
            "name": "Aave Arbitrum USDCn",
            "releases": [SULU],
            "symbol": "aArbUSDCn",
            "type": "aave-v3",
            "underlying": "0xaf88d065e77c8cC2239327C5EDb3A432268e5831",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x50834F3163758fcC1Df9973b6e91f0F0F0434aD3",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 8,
            "id": "0x078f358208685046a11C85e8ad32895DED33A249",
            "name": "Aave Arbitrum WBTC",
            "releases": [SULU],
            "symbol": "aArbWBTC",
            "type": "aave-v3",
            "underlying": "0x2f2a2543B76A4166549F7aaB2e75Bef0aefC5B0f",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xd0C7101eACbB49F3deCcCc166d238410D6D46d57",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x8437d7C167dFB82ED4Cb79CD44B7a32A1dd95c77",
            "name": "Aave Arbitrum weETH",
            "releases": [SULU],
            "symbol": "aArbweETH",
            "type": "aave-v3",
            "underlying": "0x35751007a407ca6FEFfE80b3cB397736D2cf4dbe",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xE141425bc1594b8039De6390db1cDaf4397EA22b",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x513c7E3a9c69cA3e22550eF58AC1C0088e918FFf",
            "name": "Aave Arbitrum wstETH",
            "releases": [SULU],
            "symbol": "aArbwstETH",
            "type": "aave-v3",
            "underlying": "0x5979D7b546E38E414F7E9822514be443A4800529",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb523AE262D20A936BC152e6023996e46FDC2A95D",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 6,
            "id": "0x6ab707Aca953eDAeFBc4fD23bA73294241490620",
            "name": "Aave Arbitrum USDT",
            "releases": [SULU],
            "symbol": "aArbUSDT",
            "type": "aave-v3",
            "underlying": "0xFd086bC7CD5C481DCC9C85ebE478A1C0b69FCbb9",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x3f3f5dF88dC9F13eac63DF89EC16ef6e7E25DdE7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x191c10Aa4AF7C30e871E70C95dB0E4eb77237530",
            "name": "Aave Arbitrum LINK",
            "releases": [SULU],
            "symbol": "aArbLINK",
            "type": "aave-v3",
            "underlying": "0xf97f4df75117a78c1A5a0DBb814Af92458539FB4",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb7c8Fb1dB45007F98A68Da0588e1AA524C317f27",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x6533afac2E7BCCB20dca161449A13A32D391fb00",
            "name": "Aave Arbitrum ARB",
            "releases": [SULU],
            "symbol": "aArbARB",
            "type": "aave-v3",
            "underlying": "0x912CE59144191C1204E64559FE8253a0e49E6548",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb2A824043730FE05F3DA2efaFa1CBbe83fa548D6",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0x625E7708f30cA75bfd92586e17077590C60eb4cD",
            "name": "Aave Arbitrum USDC",
            "releases": [SULU],
            "symbol": "aArbUSDC",
            "type": "aave-v3",
            "underlying": "0xFF970A61A04b1cA14834A43f5dE4533eBDDB5CC8",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x50834F3163758fcC1Df9973b6e91f0F0F0434aD3",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x82E64f49Ed5EC1bC6e43DAD4FC8Af9bb3A2312EE",
            "name": "Aave Arbitrum DAI",
            "releases": [SULU],
            "symbol": "aArbDAI",
            "type": "aave-v3",
            "underlying": "0xDA10009cBd5D07dd0CeCc66161FC93D7c9000da1",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xc5C8E77B397E531B8EC06BFb0048328B30E9eCfB",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x8Eb270e296023E9D92081fdF967dDd7878724424",
            "name": "Aave Arbitrum rETH",
            "releases": [SULU],
            "symbol": "aArbrETH",
            "type": "aave-v3",
            "underlying": "0xEC70Dcb4A1EFa46b8F2D97C310C9c4790ba5ffA8",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xD6aB2298946840262FcC278fF31516D39fF611eF",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x8ffDf2DE812095b1D19CB146E4c004587C0A0692",
            "name": "Aave Arbitrum LUSD",
            "releases": [SULU],
            "symbol": "aArbLUSD",
            "type": "aave-v3",
            "underlying": "0x93b346b6BC2548dA6A1E7d98E9a421B42541425b",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x0411D28c94d85A36bC72Cb0f875dfA8371D8fFfF",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xf329e36C7bF6E5E86ce2150875a84Ce77f477375",
            "name": "Aave Arbitrum AAVE",
            "releases": [SULU],
            "symbol": "aArbAAVE",
            "type": "aave-v3",
            "underlying": "0xba5DdD1f9d7F570dc94a51479a000E3BCE967196",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xaD1d5344AaDE45F43E596773Bcc4c423EAbdD034",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xeBe517846d0F36eCEd99C735cbF6131e1fEB775D",
            "name": "Aave Arbitrum GHO",
            "releases": [SULU],
            "symbol": "aArbGHO",
            "type": "aave-v3",
            "underlying": "0x7dfF72693f6A4149b17e7C6314655f6A9F7c8B33",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x3c786e934F23375Ca345C9b8D5aD54838796E8e7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x38d693cE1dF5AaDF7bC62595A37D667aD57922e5",
            "name": "Aave Arbitrum FRAX",
            "releases": [SULU],
            "symbol": "aArbFRAX",
            "type": "aave-v3",
            "underlying": "0x17FC002b466eEc40DaE837Fc4bE5c67993ddBd6F",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x0809E3d38d1B4214958faf06D8b1B1a2b73f2ab8",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0xe80772Eaf6e2E18B651F160Bc9158b2A5caFCA65",
            "name": "xUSD",
            "releases": [SULU],
            "symbol": "xUSD",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x6548a81E640C000150e06AB413fB3F772682e9c5",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0xFF970A61A04b1cA14834A43f5dE4533eBDDB5CC8",
            "name": "USD Coin (Arb1)",
            "releases": [SULU],
            "symbol": "USDC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x50834F3163758fcC1Df9973b6e91f0F0F0434aD3",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x17FC002b466eEc40DaE837Fc4bE5c67993ddBd6F",
            "name": "Frax",
            "releases": [SULU],
            "symbol": "FRAX",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x0809E3d38d1B4214958faf06D8b1B1a2b73f2ab8",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x5d3a1Ff2b6BAb83b63cd9AD0787074081a52ef34",
            "name": "USDe",
            "releases": [SULU],
            "symbol": "USDe",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x88AC7Bca36567525A866138F03a6F6844868E0Bc",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x211Cc4DD073734dA055fbF44a2b4667d5E5fE5d2",
            "name": "Staked USDe",
            "releases": [SULU],
            "symbol": "sUSDe",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xf2215b9c35b1697B5f47e407c917a40D055E68d7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xf7d4e7273E5015C96728A6b02f31C505eE184603",
            "name": "Staked ETH",
            "releases": [SULU],
            "symbol": "osETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xB4102D5E72c402D537C9f024F4bd9c3709FE200d",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x93b346b6BC2548dA6A1E7d98E9a421B42541425b",
            "name": "LUSD Stablecoin",
            "releases": [SULU],
            "symbol": "LUSD",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x0411D28c94d85A36bC72Cb0f875dfA8371D8fFfF",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x7dfF72693f6A4149b17e7C6314655f6A9F7c8B33",
            "name": "Gho Token",
            "releases": [SULU],
            "symbol": "GHO",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x3c786e934F23375Ca345C9b8D5aD54838796E8e7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 8,
            "id": "0x050C24dBf1eEc17babE5fc585F06116A259CC77A",
            "name": "iBTC",
            "releases": [SULU],
            "symbol": "IBTC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xd0C7101eACbB49F3deCcCc166d238410D6D46d57",
                "rate_asset": RateAsset.USD,
                "pegged_to": "WBTC",
                "non_standard": True,
            },
        },
        {
            "decimals": 6,
            "id": "0xA1b91fe9FD52141Ff8cac388Ce3F10BFDc1dE79d",
            "name": "dogwifhat",
            "releases": [],
            "symbol": "$WIF",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x25d887Ce7a35172C62FeBFD67a1856F20FaEbB00",
            "name": "Pepe",
            "releases": [],
            "symbol": "PEPE",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
    ],
)



import hashlib
import json
from web3 import AsyncWeb3
from web3.types import TContractFn, HexStr, TxParams, ChecksumAddress
from web3.contract import AsyncContract
from web3.middleware import ExtraDataToPOAMiddleware
from typing import Any


class PublicClient:
    def __init__(self, rpc_url: str):
        self.w3 = AsyncWeb3(AsyncWeb3.AsyncHTTPProvider(rpc_url))
        self.w3.middleware_onion.inject(ExtraDataToPOAMiddleware, layer=0)

        self._contracts = {}

    def contract(self, address: str, abi: list[dict[str, Any]]) -> AsyncContract:
        abi_str = json.dumps(abi, sort_keys=True)
        key_str = f"{address}_{abi_str}"
        key = hashlib.md5(key_str.encode()).hexdigest()
        if key in self._contracts:
            return self._contracts[key]

        self._contracts[key] = self.w3.eth.contract(address=address, abi=abi)
        return self._contracts[key]

    async def populated_transaction(
        self,
        function: TContractFn,
        account: ChecksumAddress,
        value: int | None = None,
    ) -> TxParams:
        nonce = await self.w3.eth.get_transaction_count(account)

        transaction_payload = {
            "nonce": nonce,
            "from": account,
        }

        if value is not None:
            transaction_payload["value"] = value

        return await function.build_transaction(transaction_payload)


class WalletClient(PublicClient):
    def __init__(
        self,
        rpc_url: str,
        private_key: str,
    ):
        super().__init__(rpc_url)
        self.account = self.w3.eth.account.from_key(private_key)

    async def populated_transaction(
        self,
        function: TContractFn,
        account: ChecksumAddress | None = None,
        value: int | None = None,
    ) -> TxParams:
        return await super().populated_transaction(
            function, account or self.account.address, value
        )

    async def send_transaction(self, transaction: TxParams) -> HexStr:
        signed_transaction = self.w3.eth.account.sign_transaction(
            transaction, self.account.key
        )
        tx_hash = await self.w3.eth.send_raw_transaction(
            signed_transaction.raw_transaction
        )
        return tx_hash.to_0x_hex()



from web3 import Web3
from web3.types import ChecksumAddress, HexStr, TxParams
from eth_abi import encode
from ...utils.clients import PublicClient, WalletClient
from ...utils.encoding import encoding_to_types
from ....abis import abis


# --------------------------------------------------------------------------------------------
# FACTORY
# --------------------------------------------------------------------------------------------


def encode_deploy_proxy_construct_data(
    owner: ChecksumAddress,
    borrowed_assets_recipient: ChecksumAddress,
) -> HexStr:
    contract = Web3().eth.contract(abi=abis.IAaveV3FlashLoanAssetManager)
    return contract.encode_abi(
        "init",
        [owner, borrowed_assets_recipient],
    )


async def deploy_proxy(
    client: WalletClient,
    dispatcher_owned_beacon_factory: ChecksumAddress,
    construct_data: HexStr,
) -> TxParams:
    construct_data = Web3.to_bytes(hexstr=construct_data)
    contract = client.contract(
        dispatcher_owned_beacon_factory, abis.IDispatcherOwnedBeaconFactory
    )
    function = contract.functions.deployProxy(construct_data)
    return await client.populated_transaction(function)


# --------------------------------------------------------------------------------------------
# LIB
# --------------------------------------------------------------------------------------------


FLASH_LOAN_CALLS_ENCODING = [
    {
        "type": "tuple[]",
        "name": "calls",
        "components": [
            {
                "name": "target",
                "type": "address",
            },
            {
                "name": "data",
                "type": "bytes",
            },
        ],
    },
]


def encode_flash_loan_calls(
    calls: list[ChecksumAddress, HexStr],
) -> HexStr:
    types = encoding_to_types(FLASH_LOAN_CALLS_ENCODING)
    args = [[[target, Web3.to_bytes(hexstr=data)] for target, data in calls]]
    return Web3.to_hex(encode(types, args))


async def flash_loan(
    client: WalletClient,
    aave_v3_flash_loan_asset_manager: ChecksumAddress,
    assets: list[ChecksumAddress],
    amounts: list[int],
    encoded_calls: HexStr,
) -> TxParams:
    contract = client.contract(
        aave_v3_flash_loan_asset_manager, abis.IAaveV3FlashLoanAssetManager
    )
    function = contract.functions.flashLoan(assets, amounts, encoded_calls)
    return await client.populated_transaction(function)


async def get_borrowed_assets_recipient(
    client: PublicClient,
    aave_v3_flash_loan_asset_manager: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(
        aave_v3_flash_loan_asset_manager, abis.IAaveV3FlashLoanAssetManager
    )
    function = contract.functions.getBorrowedAssetsRecipient()
    return await function.call()


async def get_owner(
    client: PublicClient,
    aave_v3_flash_loan_asset_manager: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(
        aave_v3_flash_loan_asset_manager, abis.IAaveV3FlashLoanAssetManager
    )
    function = contract.functions.getOwner()
    return await function.call()



from typing import Any, Literal, TypedDict, get_args
from web3.types import ChecksumAddress
from .assets import PrimitiveAsset
from .types import Network


# Network = Literal[42161, 8453, 1, 137] -> defined in types.py to avoid circular import


NetworkSlug = Literal["arbitrum", "base", "ethereum", "polygon"]


class NetworkDefinitionCurrency(TypedDict):
    wrapper: ChecksumAddress
    native_token: PrimitiveAsset


class NetworkDefinitionExplorer(TypedDict):
    label: str
    url: str


class NetworkDefinition(TypedDict):
    currency: NetworkDefinitionCurrency
    explorer: NetworkDefinitionExplorer
    id: Network
    slug: NetworkSlug
    label: str
    rpc: str


def get_network(network_or_slug: Network | NetworkSlug) -> NetworkDefinition:
    if is_supported_network(network_or_slug):
        return NETWORKS[network_or_slug]
    if is_supported_network_slug(network_or_slug):
        return NETWORKS[NETWORK_BY_SLUG[network_or_slug]]
    raise ValueError(f"Invalid network {network_or_slug}")


def is_network_identifier(value: Any) -> bool:
    return is_supported_network(value) or is_supported_network_slug(value)


def is_supported_network_slug(value: Any) -> bool:
    return isinstance(value, str) and value in get_args(NetworkSlug)


def is_supported_network(value: Any) -> bool:
    return isinstance(value, int) and value in get_args(Network)


SLUG_BY_NETWORK: dict[Network, NetworkSlug] = dict(zip(get_args(Network), get_args(NetworkSlug)))


NETWORK_BY_SLUG: dict[NetworkSlug, Network] = dict(zip(get_args(NetworkSlug), get_args(Network)))


ARBITRUM: NetworkDefinition = {
    "currency": {
        "wrapper": "0x82aF49447D8a07e3bd95BD0d56f35241523fBab1",
        "native_token": {
            "id": "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "name": "Ether",
            "symbol": "ETH",
            "decimals": 18,
            "type": "primitive",  # AssetType from .assets
            "releases": [],
            "network": NETWORK_BY_SLUG["arbitrum"],
            "registered": False,
            "price_feed": {
                "type": "NONE",  # PriceFeedType from .price_feeds
            },
        },
    },
    "explorer": {
        "label": "Arbiscan",
        "url": "https://arbiscan.io/",
    },
    "id": NETWORK_BY_SLUG["arbitrum"],
    "label": "Arbitrum",
    "rpc": "https://arb1.arbitrum.io/rpc",
    "slug": "arbitrum",  # NetworkSlug
}


BASE: NetworkDefinition = {
    "currency": {
        "wrapper": "0x4200000000000000000000000000000000000006",
        "native_token": {
            "id": "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "name": "Ether",
            "symbol": "ETH",
            "decimals": 18,
            "type": "primitive",  # AssetType from .assets
            "releases": [],
            "network": NETWORK_BY_SLUG["base"],
            "registered": False,
            "price_feed": {
                "type": "NONE",  # PriceFeedType from .price_feeds
            },
        },
    },
    "explorer": {
        "label": "BaseScan",
        "url": "https://basescan.org/",
    },
    "id": NETWORK_BY_SLUG["base"],
    "label": "Base",
    "rpc": "https://mainnet.base.org",
    "slug": "base",  # NetworkSlug
}


MAINNET: NetworkDefinition = {
    "currency": {
        "wrapper": "0xC02aaA39b223FE8D0A0e5C4F27eAD9083C756Cc2",
        "native_token": {
            "id": "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE",
            "name": "Ether",
            "symbol": "ETH",
            "decimals": 18,
            "type": "primitive",  # AssetType from .assets
            "releases": [],
            "network": NETWORK_BY_SLUG["ethereum"],
            "registered": False,
            "price_feed": {
                "type": "NONE",  # PriceFeedType from .price_feeds
            },
        },
    },
    "explorer": {
        "label": "Etherscan",
        "url": "https://etherscan.io",
    },
    "id": NETWORK_BY_SLUG["ethereum"],
    "label": "Ethereum",
    "rpc": "https://mainnet.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161",
    "slug": "ethereum",  # NetworkSlug
}


POLYGON: NetworkDefinition = {
    "currency": {
        "wrapper": "0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270",
        "native_token": {
            "id": "0x0000000000000000000000000000000000001010",
            "name": "Polygon Ecosystem Token",
            "symbol": "POL",
            "decimals": 18,
            "type": "primitive",  # AssetType from .assets
            "releases": [],
            "network": NETWORK_BY_SLUG["polygon"],
            "registered": False,
            "price_feed": {
                "type": "NONE",  # PriceFeedType from .price_feeds
            },
        },
    },
    "explorer": {
        "label": "Polygonscan",
        "url": "https://polygonscan.com",
    },
    "id": NETWORK_BY_SLUG["polygon"],
    "label": "Polygon",
    "rpc": "https://polygon-rpc.com",
    "slug": "polygon",  # NetworkSlug
}


NETWORKS = dict(zip(get_args(Network), [ARBITRUM, BASE, MAINNET, POLYGON]))



from web3.types import ChecksumAddress
from .utils.clients import PublicClient
from ..abis import abis


async def is_supported_asset(
    client: PublicClient,
    value_interpreter: ChecksumAddress,
    asset: ChecksumAddress,
) -> bool:
    contract = client.contract(value_interpreter, abis.IValueInterpreter)
    function = contract.functions.isSupportedAsset(asset)
    return await function.call()


async def is_supported_asset_for_derivative_price_feed(
    client: PublicClient,
    derivative_price_feed: ChecksumAddress,
    asset: ChecksumAddress,
) -> bool:
    contract = client.contract(derivative_price_feed, abis.IDerivativePriceFeed)
    function = contract.functions.isSupportedAsset(asset)
    return await function.call()


async def is_supported_primitive_asset(
    client: PublicClient,
    value_interpreter: ChecksumAddress,
    asset: ChecksumAddress,
) -> bool:
    contract = client.contract(value_interpreter, abis.IValueInterpreter)
    function = contract.functions.isSupportedPrimitiveAsset(asset)
    return await function.call()


async def is_supported_derivative_asset(
    client: PublicClient,
    value_interpreter: ChecksumAddress,
    asset: ChecksumAddress,
) -> bool:
    contract = client.contract(value_interpreter, abis.IValueInterpreter)
    function = contract.functions.isSupportedDerivativeAsset(asset)
    return await function.call()


async def get_aggregator_for_primitive(
    client: PublicClient,
    value_interpreter: ChecksumAddress,
    asset: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(value_interpreter, abis.IValueInterpreter)
    function = contract.functions.getAggregatorForPrimitive(asset)
    return await function.call()


async def get_rate_asset_for_primitive(
    client: PublicClient,
    value_interpreter: ChecksumAddress,
    asset: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(value_interpreter, abis.IValueInterpreter)
    function = contract.functions.getRateAssetForPrimitive(asset)
    return await function.call()


async def get_price_feed_for_derivative(
    client: PublicClient,
    value_interpreter: ChecksumAddress,
    asset: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(value_interpreter, abis.IValueInterpreter)
    function = contract.functions.getPriceFeedForDerivative(asset)
    return await function.call()


async def get_eth_usd_aggregator(
    client: PublicClient,
    value_interpreter: ChecksumAddress,
) -> ChecksumAddress:
    contract = client.contract(value_interpreter, abis.IValueInterpreter)
    function = contract.functions.getEthUsdAggregator()
    return await function.call()



from typing import Literal


Network = Literal[42161, 8453, 1, 137]


Release = Literal[
    "arbitrum.sulu",
    "base.sulu",
    "ethereum.sulu",
    "ethereum.encore",
    "ethereum.phoenix",
    "polygon.sulu",
    "testnet.sulu",
]


from .adapters import *
from .assets import *
from .contracts import *
from .currencies import *
from .environment import *
from .exchanges import *
from .networks import *
from .price_feeds import *
from .protocols import *
from .releases import *
from .utils import *


from typing import TypedDict
from web3.types import ChecksumAddress, HexStr, TxParams
from ..utils.clients import WalletClient
from ...abis import abis


class CallExtensionParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    extension_manager: ChecksumAddress
    action_id: int
    call_args: HexStr


async def call_extension(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
    extension_manager: ChecksumAddress,
    action_id: int,
    call_args: HexStr,
) -> TxParams:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.callOnExtension(
        extension_manager, action_id, call_args
    )
    return await client.populated_transaction(function)



from web3 import Web3
from web3.constants import ADDRESS_ZERO
from eth_typing import ChecksumAddress
from ..contracts import Version, is_version
from ..environment import Environment, EnvironmentGroup
from ..releases import Deployment, Release, is_deployment
from ..networks import Network, NetworkSlug, get_network, is_network_identifier
from .arbitrum import DEPLOYMENT as ARBITRUM
from .base import DEPLOYMENT as BASE
from .ethereum import DEPLOYMENT as ETHEREUM
from .polygon import DEPLOYMENT as POLYGON
from .testnet import DEPLOYMENT as TESTNET


DEPLOYMENTS = {
    "arbitrum": ARBITRUM,
    "base": BASE,
    "ethereum": ETHEREUM,
    "polygon": POLYGON,
    "testnet": TESTNET,
}


def get_environment_group(deployment: Deployment) -> EnvironmentGroup:
    return EnvironmentGroup(DEPLOYMENTS[deployment])


def get_environment_for_release(release: Release) -> Environment:
    deployment, version = release.split(".")
    if is_deployment(deployment) and is_version(version):
        return get_environment(deployment, version)
    raise ValueError(f"Unknown release {release}")


def get_environment(
    deployment_or_network: Deployment | Network | NetworkSlug,
    version_or_address: Version | ChecksumAddress,
) -> Environment:
    if is_deployment(deployment_or_network):
        deployment = DEPLOYMENTS[deployment_or_network]
    elif is_network_identifier(deployment_or_network):
        network = get_network(deployment_or_network)
        deployment = next(
            (
                deployment
                for deployment in DEPLOYMENTS.values()
                if deployment["kind"] == "live"
                and deployment["network"] == network["id"]
            ),
            None,
        )

        if deployment is None:
            raise ValueError(f"Failed to find deployment for network {network['slug']}")

    else:
        raise ValueError(f"Failed to find deployment {deployment_or_network}")

    if is_version(version_or_address):
        release = deployment["releases"].get(version_or_address)

        if release is not None:
            return Environment(deployment, release["version"])
    elif (
        Web3.is_checksum_address(version_or_address)
        and version_or_address != ADDRESS_ZERO
    ):
        release = next(
            (
                release
                for release in deployment["releases"].values()
                if release["address"] == version_or_address
            ),
            None,
        )
        if release is not None:
            return Environment(deployment, release["version"])

    raise ValueError(
        f"Failed to find {version_or_address} release on deployment {deployment['slug']}"
    )



from web3.types import ChecksumAddress
from ..configuration import get_enabled_policies
from ..utils.clients import PublicClient


async def is_enabled(
    client: PublicClient,
    policy: ChecksumAddress,
    policy_manager: ChecksumAddress,
    comptroller_proxy: ChecksumAddress,
) -> bool:
    enabled_policies = await get_enabled_policies(
        client, policy_manager, comptroller_proxy
    )
    return policy in enabled_policies



from ..releases import RELEASES
from ..assets import define_asset_list
from ..networks import NETWORK_BY_SLUG
from ..price_feeds import RateAsset


TESTNET, POLYGON = [RELEASES[r] for r in ["testnet", "polygon"]]


ASSETS = define_asset_list(
    NETWORK_BY_SLUG["polygon"],
    [
        {
            "decimals": 18,
            "id": "0x00e5646f60AC6Fb446f621d146B6E1886f002905",
            "name": "Rai Reflex Index (PoS)",
            "releases": [],
            "symbol": "RAI",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x0297e37f1873D2DAb4487Aa67cD56B58E2F27875",
            "name": "Balancer Polygon Base Pool",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "B-POLYBASE",
            "staking": [
                {
                    "token": "0x068Ff98072d3eB848D012e3390703BB507729ed6",
                    type: "gauge",
                }
            ],
            "pool_id": "0x0297e37f1873d2dab4487aa67cd56b58e2f27875000100000000000000000002",
            "pool_factory": "0x8E9aa87E45e92bad84D5F8DD1bff34Fb92637dE9",
            "pool_type": "weighted",
            "underlyings": [
                "0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270",
                "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
                "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
                "0x9a71012B13CA4d3D0Cdc72A177DF3ef03b0E76A3",
            ],
            "type": "balancer-pool",
            "price_feed": {
                "type": "DERIVATIVE_BALANCER_V2_WEIGHTED_POOL",
                "address": "0x30ee2560675f8A523cD07B109FcB472a54b08314",
            },
        },
        {
            "decimals": 18,
            "id": "0x068Ff98072d3eB848D012e3390703BB507729ed6",
            "name": "Balancer B-POLYBASE RewardGauge Deposit",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "B-POLYBASE-gauge",
            "pool": "0x0297e37f1873D2DAb4487Aa67cD56B58E2F27875",
            "pool_type": "weighted",
            "gauge_factory": "0x3b8cA519122CdD8efb272b0D3085453404B25bD0",
            "underlyings": [
                "0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270",
                "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
                "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
                "0x9a71012B13CA4d3D0Cdc72A177DF3ef03b0E76A3",
            ],
            "type": "balancer-pool-gauge",
            "price_feed": {
                "type": "DERIVATIVE_BALANCER_V2_GAUGE_TOKEN",
                "address": "0xBd19713E7fE01d3958afbB4fB5b3802a0d53605c",
            },
        },
        {
            "decimals": 18,
            "id": "0x3d468AB2329F296e1b9d8476Bb54Dd77D8c2320f",
            "name": "20WETH-80BAL",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "staking": [],
            "symbol": "20WETH-80BAL",
            "pool_id": "0x3d468ab2329f296e1b9d8476bb54dd77d8c2320f000200000000000000000426",
            "pool_factory": "0x8E9aa87E45e92bad84D5F8DD1bff34Fb92637dE9",
            "pool_type": "weighted",
            "underlyings": [
                "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
                "0x9a71012B13CA4d3D0Cdc72A177DF3ef03b0E76A3",
            ],
            "type": "balancer-pool",
            "price_feed": {
                "type": "DERIVATIVE_BALANCER_V2_WEIGHTED_POOL",
                "address": "0x30ee2560675f8A523cD07B109FcB472a54b08314",
            },
        },
        {
            "decimals": 18,
            "id": "0x03b54A6e9a984069379fae1a4fC4dBAE93B3bCCD",
            "name": "Wrapped liquid staked Ether 2.0 (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "wstETH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x10f964234cae09cB6a9854B56FF7D4F38Cda5E6a",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xf59036CAEBeA7dC4b86638DFA2E3C97dA9FcCd40",
            "name": "Aave Polygon wstETH",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolwstETH",
            "type": "aave-v3",
            "underlying": "0x03b54A6e9a984069379fae1a4fC4dBAE93B3bCCD",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x10f964234cae09cB6a9854B56FF7D4F38Cda5E6a",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x03cD191F589d12b0582a99808cf19851E468E6B5",
            "name": "Balancer Polygon Tricrypto",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "BPTC",
            "staking": [
                {
                    "token": "0x0F09F70Ed59202c77aC667f574A5f79bC65CeA48",
                    type: "gauge",
                }
            ],
            "pool_id": "0x03cd191f589d12b0582a99808cf19851e468e6b500010000000000000000000a",
            "pool_factory": "0x8E9aa87E45e92bad84D5F8DD1bff34Fb92637dE9",
            "pool_type": "weighted",
            "underlyings": [
                "0x1BFD67037B42Cf73acF2047067bd4F2C47D9BfD6",
                "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
                "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
            ],
            "type": "balancer-pool",
            "price_feed": {
                "type": "DERIVATIVE_BALANCER_V2_WEIGHTED_POOL",
                "address": "0x30ee2560675f8A523cD07B109FcB472a54b08314",
            },
        },
        {
            "decimals": 18,
            "id": "0x0F09F70Ed59202c77aC667f574A5f79bC65CeA48",
            "name": "Balancer BPTC RewardGauge Deposit",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "BPTC-gauge",
            "pool": "0x03cD191F589d12b0582a99808cf19851E468E6B5",
            "pool_type": "weighted",
            "gauge_factory": "0x3b8cA519122CdD8efb272b0D3085453404B25bD0",
            "underlyings": [
                "0x1BFD67037B42Cf73acF2047067bd4F2C47D9BfD6",
                "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
                "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
            ],
            "type": "balancer-pool-gauge",
            "price_feed": {
                "type": "DERIVATIVE_BALANCER_V2_GAUGE_TOKEN",
                "address": "0xBd19713E7fE01d3958afbB4fB5b3802a0d53605c",
            },
        },
        {
            "decimals": 18,
            "id": "0x36128D5436d2d70cab39C9AF9CcE146C38554ff0",
            "name": "Balancer Polygon DeFi Index",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "B-POLYDEFI",
            "staking": [
                {
                    "token": "0xe32AF68E3e23dd01441f598B1D967A0bF7c6B407",
                    type: "gauge",
                }
            ],
            "pool_id": "0x36128d5436d2d70cab39c9af9cce146c38554ff0000100000000000000000008",
            "pool_factory": "0x8E9aa87E45e92bad84D5F8DD1bff34Fb92637dE9",
            "pool_type": "weighted",
            "underlyings": [
                "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
                "0x53E0bca35eC356BD5ddDFebbD1Fc0fD03FaBad39",
                "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
                "0x9a71012B13CA4d3D0Cdc72A177DF3ef03b0E76A3",
                "0xD6DF932A45C0f255f85145f286eA0b292B21C90B",
            ],
            "type": "balancer-pool",
            "price_feed": {
                "type": "DERIVATIVE_BALANCER_V2_WEIGHTED_POOL",
                "address": "0x30ee2560675f8A523cD07B109FcB472a54b08314",
            },
        },
        {
            "decimals": 18,
            "id": "0xe32AF68E3e23dd01441f598B1D967A0bF7c6B407",
            "name": "Balancer B-POLYDEFI RewardGauge Deposit",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "B-POLYDEFI-gauge",
            "pool": "0x36128D5436d2d70cab39C9AF9CcE146C38554ff0",
            "pool_type": "weighted",
            "gauge_factory": "0x3b8cA519122CdD8efb272b0D3085453404B25bD0",
            "underlyings": [
                "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
                "0x53E0bca35eC356BD5ddDFebbD1Fc0fD03FaBad39",
                "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
                "0x9a71012B13CA4d3D0Cdc72A177DF3ef03b0E76A3",
                "0xD6DF932A45C0f255f85145f286eA0b292B21C90B",
            ],
            "type": "balancer-pool-gauge",
            "price_feed": {
                "type": "DERIVATIVE_BALANCER_V2_GAUGE_TOKEN",
                "address": "0xBd19713E7fE01d3958afbB4fB5b3802a0d53605c",
            },
        },
        {
            "decimals": 18,
            "id": "0x3A58a54C066FdC0f2D55FC9C89F0415C92eBf3C4",
            "name": "Staked MATIC (PoS)",
            "symbol": "stMATIC",
            "type": "primitive",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x97371dF4492605486e23Da797fA68e55Fc38a13f",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xEA1132120ddcDDA2F119e99Fa7A27a0d036F7Ac9",
            "name": "Aave Polygon STMATIC",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolSTMATIC",
            "type": "aave-v3",
            "underlying": "0x3A58a54C066FdC0f2D55FC9C89F0415C92eBf3C4",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x97371dF4492605486e23Da797fA68e55Fc38a13f",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x080b5BF8f360F624628E0fb961F4e67c9e3c7CF1",
            "name": "Aave GHST",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amGHST",
            "type": "aave-v2",
            "underlying": "0x385Eeac5cB85A38A9a07A70c73e0a3271CfB54A7",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xe638249AF9642CdA55A92245525268482eE4C67b",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x8Eb270e296023E9D92081fdF967dDd7878724424",
            "name": "Aave Polygon GHST",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolGHST",
            "type": "aave-v3",
            "underlying": "0x385Eeac5cB85A38A9a07A70c73e0a3271CfB54A7",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xe638249AF9642CdA55A92245525268482eE4C67b",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x0b3F868E0BE5597D5DB7fEB59E1CADBb0fdDa50a",
            "name": "SushiToken (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "SUSHI",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x17414Eb5159A082e8d41D243C1601c2944401431",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x0Ca2e42e8c21954af73Bc9af1213E4e81D6a669A",
            "name": "Aave LINK",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amLINK",
            "type": "aave-v2",
            "underlying": "0x53E0bca35eC356BD5ddDFebbD1Fc0fD03FaBad39",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb77fa460604b9C6435A235D057F7D319AC83cb53",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x191c10Aa4AF7C30e871E70C95dB0E4eb77237530",
            "name": "Aave Polygon LINK",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolLINK",
            "type": "aave-v3",
            "underlying": "0x53E0bca35eC356BD5ddDFebbD1Fc0fD03FaBad39",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb77fa460604b9C6435A235D057F7D319AC83cb53",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270",
            "name": "Wrapped Polygon Ecosystem Token",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "WPOL",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xAB594600376Ec9fD91F8e885dADF0CE036862dE0",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x104592a158490a9228070E0A8e5343B499e125D0",
            "name": "Frax (PoS)",
            "releases": [],
            "symbol": "FRAX",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x172370d5Cd63279eFa6d502DAB29171933a610AF",
            "name": "CRV (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "CRV",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x1CF68C76803c9A415bE301f50E82e44c64B7F1D4",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x18ec0A6E18E5bc3784fDd3a3634b31245ab704F6",
            "name": "Monerium EUR emoney",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "EURe",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x73366Fe0AA0Ded304479862808e02506FE556a98",
                "rate_asset": RateAsset.USD,
                "pegged_to": "EUR",
                "non_standard": True,
            },
        },
        {
            "decimals": 6,
            "id": "0x1a13F4Ca1d028320A707D99520AbFefca3998b7F",
            "name": "Aave USDC",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amUSDC",
            "type": "aave-v2",
            "underlying": "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xfE4A8cc5b5B2366C1B58Bea3858e81843581b2F7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0x625E7708f30cA75bfd92586e17077590C60eb4cD",
            "name": "Aave Polygon USDC",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolUSDC",
            "type": "aave-v3",
            "underlying": "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xfE4A8cc5b5B2366C1B58Bea3858e81843581b2F7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x1a3acf6D19267E2d3e7f898f42803e90C9219062",
            "name": "Frax Share",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "FXS",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x6C0fe985D3cAcbCdE428b84fc9431792694d0f51",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 8,
            "id": "0x1BFD67037B42Cf73acF2047067bd4F2C47D9BfD6",
            "name": "(PoS) Wrapped BTC",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "WBTC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xA338e0492B2F944E9F8C0653D3AD1484f2657a37",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x1C954E8fe737F99f68Fa1CCda3e51ebDB291948C",
            "name": "Kyber Network Crystal v2 (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "KNC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x10e5f3DFc81B3e5Ef4e648C4454D04e79E1E41E2",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x1d2a0E5EC8E5bBDCA5CB219e649B565d8e5c3360",
            "name": "Aave AAVE",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amAAVE",
            "type": "aave-v2",
            "underlying": "0xD6DF932A45C0f255f85145f286eA0b292B21C90B",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xbE23a3AA13038CfC28aFd0ECe4FdE379fE7fBfc4",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xf329e36C7bF6E5E86ce2150875a84Ce77f477375",
            "name": "Aave Polygon AAVE",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolAAVE",
            "type": "aave-v3",
            "underlying": "0xD6DF932A45C0f255f85145f286eA0b292B21C90B",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xbE23a3AA13038CfC28aFd0ECe4FdE379fE7fBfc4",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x1FcbE5937B0cc2adf69772D228fA4205aCF4D9b2",
            "name": "Badger (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "BADGER",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x82C9d4E88862f194C2bd874a106a90dDD0D35AAB",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x21eC9431B5B55c5339Eb1AE7582763087F98FAc2",
            "name": "Aave SUSHI",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amSUSHI",
            "type": "aave-v2",
            "underlying": "0x0b3F868E0BE5597D5DB7fEB59E1CADBb0fdDa50a",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x17414Eb5159A082e8d41D243C1601c2944401431",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xc45A479877e1e9Dfe9FcD4056c699575a1045dAA",
            "name": "Aave Polygon SUSHI",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolSUSHI",
            "type": "aave-v3",
            "underlying": "0x0b3F868E0BE5597D5DB7fEB59E1CADBb0fdDa50a",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x17414Eb5159A082e8d41D243C1601c2944401431",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x24834BBEc7E39ef42f4a75EAF8E5B6486d3F0e57",
            "name": "Wrapped LUNA Token (PoS)",
            "releases": [],
            "symbol": "LUNA",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 6,
            "id": "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
            "name": "USD Coin (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "USDC.e",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xfE4A8cc5b5B2366C1B58Bea3858e81843581b2F7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x27F8D03b3a2196956ED754baDc28D73be8830A6e",
            "name": "Aave DAI",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amDAI",
            "type": "aave-v2",
            "underlying": "0x8f3Cf7ad23Cd3CaDbD9735AFf958023239c6A063",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x4746DeC9e833A82EC7C2C1356372CcF2cfcD2F3D",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x82E64f49Ed5EC1bC6e43DAD4FC8Af9bb3A2312EE",
            "name": "Aave Polygon DAI",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolDAI",
            "type": "aave-v3",
            "underlying": "0x8f3Cf7ad23Cd3CaDbD9735AFf958023239c6A063",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x4746DeC9e833A82EC7C2C1356372CcF2cfcD2F3D",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x28424507fefb6f7f8E9D3860F56504E4e5f5f390",
            "name": "Aave WETH",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amWETH",
            "type": "aave-v2",
            "underlying": "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xF9680D99D6C9589e2a93a78A04A279e509205945",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xe50fA9b3c56FfB159cB0FCA61F5c9D750e8128c8",
            "name": "Aave Polygon WETH",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolWETH",
            "type": "aave-v3",
            "underlying": "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xF9680D99D6C9589e2a93a78A04A279e509205945",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xE5417Af564e4bFDA1c483642db72007871397896",
            "name": "Gains Network",
            "releases": [POLYGON["sulu"]],
            "symbol": "GNS",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x9cb43aa3D036Cb035a694Ba0AAa91f8875B16cE1",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x2b88aD57897A8b496595925F43048301C37615Da",
            "name": "PickleToken (PoS)",
            "releases": [],
            "symbol": "PICKLE",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x3066818837c5e6eD6601bd5a91B0762877A6B731",
            "name": "UMA Voting Token v1 (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "UMA",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x33D9B1BAaDcF4b26ab6F8E83e9cb8a611B2B3956",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x308ba429a913d952275090Fc35dEA9C36A511B5A",
            "name": "Wrapped Tezos (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "WXTZ",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x691e26AB58ff05800E028b0876A41B720b26FC65",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x385Eeac5cB85A38A9a07A70c73e0a3271CfB54A7",
            "name": "Aavegotchi GHST Token (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "GHST",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xe638249AF9642CdA55A92245525268482eE4C67b",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x3962F4A0A0051DccE0be73A7e09cEf5756736712",
            "name": "Livepeer Token (PoS)",
            "releases": [],
            "symbol": "LPT",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x3AE490db48d74B1bC626400135d4616377D0109f",
            "name": "AlphaToken (PoS)",
            "releases": [],
            "symbol": "ALPHA",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x3BA4c387f786bFEE076A58914F5Bd38d668B42c3",
            "name": "BNB (PoS)",
            "releases": [],
            "symbol": "BNB",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x3Cef98bb43d732E2F285eE605a8158cDE967D219",
            "name": "Basic Attention Token (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "BAT",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x2346Ce62bd732c62618944E51cbFa09D985d86D2",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0x3c499c542cEF5E3811e1192ce70d8cC03d5c3359",
            "name": "USD Coin",
            "releases": [POLYGON["sulu"]],
            "symbol": "USDC",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xfE4A8cc5b5B2366C1B58Bea3858e81843581b2F7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x3Df8f92b7E798820ddcCA2EBEA7BAbda2c90c4aD",
            "name": "Aave CRV",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amCRV",
            "type": "aave-v2",
            "underlying": "0x172370d5Cd63279eFa6d502DAB29171933a610AF",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x1CF68C76803c9A415bE301f50E82e44c64B7F1D4",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x513c7E3a9c69cA3e22550eF58AC1C0088e918FFf",
            "name": "Aave Polygon CRV",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolCRV",
            "type": "aave-v3",
            "underlying": "0x172370d5Cd63279eFa6d502DAB29171933a610AF",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x1CF68C76803c9A415bE301f50E82e44c64B7F1D4",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x3e121107F6F22DA4911079845a470757aF4e1A1b",
            "name": "Frax Share (PoS)",
            "releases": [],
            "symbol": "FXS",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x41b3966B4FF7b427969ddf5da3627d6AEAE9a48E",
            "name": "Nexo (PoS)",
            "releases": [],
            "symbol": "NEXO",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x4257EA7637c355F81616050CbB6a9b709fd72683",
            "name": "Convex Token (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "CVX",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x5ec151834040B4D453A1eA46aA634C1773b36084",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x45c32fA6DF82ead1e2EF74d17b76547EDdFaFF89",
            "name": "Frax",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "FRAX",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x00DBeB1e45485d53DF7C2F0dF1Aa0b6Dc30311d3",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x50B728D8D964fd00C2d0AAD81718b71311feF68a",
            "name": "Synthetix Network Token (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "SNX",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xbF90A5D9B6EE9019028dbFc2a9E50056d5252894",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x53E0bca35eC356BD5ddDFebbD1Fc0fD03FaBad39",
            "name": "ChainLink Token",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "LINK",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xb77fa460604b9C6435A235D057F7D319AC83cb53",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x5559Edb74751A0edE9DeA4DC23aeE72cCA6bE3D5",
            "name": "0x Protocol Token (PoS)",
            "releases": [],
            "symbol": "ZRX",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 8,
            "id": "0x5c2ed810328349100A66B82b78a1791B101C9D61",
            "name": "Aave WBTC",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amWBTC",
            "type": "aave-v2",
            "underlying": "0x1BFD67037B42Cf73acF2047067bd4F2C47D9BfD6",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xA338e0492B2F944E9F8C0653D3AD1484f2657a37",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 8,
            "id": "0x078f358208685046a11C85e8ad32895DED33A249",
            "name": "Aave Polygon WBTC",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolWBTC",
            "type": "aave-v3",
            "underlying": "0x1BFD67037B42Cf73acF2047067bd4F2C47D9BfD6",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xA338e0492B2F944E9F8C0653D3AD1484f2657a37",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x5fe2B58c013d7601147DcdD68C143A77499f5531",
            "name": "Graph Token (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "GRT",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x3FabBfb300B1e2D7c9B84512fe9D30aeDF24C410",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0x60D55F02A771d515e077c9C2403a1ef324885CeC",
            "name": "Aave USDT",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amUSDT",
            "type": "aave-v2",
            "underlying": "0xc2132D05D31c914a87C6611C10748AEb04B58e8F",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x0A6513e40db6EB1b165753AD52E80663aeA50545",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0x6ab707Aca953eDAeFBc4fD23bA73294241490620",
            "name": "Aave Polygon USDT",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolUSDT",
            "type": "aave-v3",
            "underlying": "0xc2132D05D31c914a87C6611C10748AEb04B58e8F",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x0A6513e40db6EB1b165753AD52E80663aeA50545",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x6811079E3c63ED96Eb005384d7E7ec8810E3D521",
            "name": "SushiBar (PoS)",
            "releases": [],
            "symbol": "xSUSHI",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x6f7C932e7684666C9fd1d44527765433e01fF61d",
            "name": "MAKER (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "MKR",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x807b59d12520830D1864286FA0271c27baa94197",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
            "name": "Wrapped Ether",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "WETH",
            "type": "primitive",
            "price_feed": {
                "type": "WETH",
            },
        },
        {
            "decimals": 18,
            "id": "0x7eC26842F195c852Fa843bB9f6D8B583a274a157",
            "name": "Enjin Coin (PoS)",
            "releases": [],
            "symbol": "ENJ",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x81fB82aAcB4aBE262fc57F06fD4c1d2De347D7B1",
            "name": "Aave DPI",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amDPI",
            "type": "aave-v2",
            "underlying": "0x85955046DF4668e1DD369D2DE9f3AEB98DD2A369",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xC70aAF9092De3a4E5000956E672cDf5E996B4610",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x724dc807b04555b71ed48a6896b6F41593b8C637",
            "name": "Aave Polygon DPI",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolDPI",
            "type": "aave-v3",
            "underlying": "0x85955046DF4668e1DD369D2DE9f3AEB98DD2A369",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xC70aAF9092De3a4E5000956E672cDf5E996B4610",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x831753DD7087CaC61aB5644b308642cc1c33Dc13",
            "name": "Quickswap",
            "releases": [],
            "symbol": "QUICK",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x8505b9d2254A7Ae468c0E9dd10Ccea3A837aef5c",
            "name": "(PoS) Compound",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "COMP",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x2A8758b7257102461BC958279054e372C2b1bDE6",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x85955046DF4668e1DD369D2DE9f3AEB98DD2A369",
            "name": "DefiPulse Index (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "DPI",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xC70aAF9092De3a4E5000956E672cDf5E996B4610",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x8dF3aad3a84da6b69A4DA8aeC3eA40d9091B2Ac4",
            "name": "Aave WMATIC",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amWMATIC",
            "type": "aave-v2",
            "underlying": "0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xAB594600376Ec9fD91F8e885dADF0CE036862dE0",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x6d80113e533a2C0fe82EaBD35f1875DcEA89Ea97",
            "name": "Aave Polygon WMATIC",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolWMATIC",
            "type": "aave-v3",
            "underlying": "0x0d500B1d8E8eF31E21C99d1Db9A6444d3ADf1270",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xAB594600376Ec9fD91F8e885dADF0CE036862dE0",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x8f3Cf7ad23Cd3CaDbD9735AFf958023239c6A063",
            "name": "(PoS) Dai Stablecoin",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "DAI",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x4746DeC9e833A82EC7C2C1356372CcF2cfcD2F3D",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x918C749A09fA539E1Dc8019F875fab9EB1EBc6A3",
            "name": "FTT (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "FTT",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x817A7D43f0277Ca480AE03Ec76Fc63A2EC7114bA",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x95c300e7740D2A88a44124B424bFC1cB2F9c3b89",
            "name": "Alchemix (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "ALCX",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x5DB6e61B6159B20F068dc15A47dF2E5931b14f29",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x9a71012B13CA4d3D0Cdc72A177DF3ef03b0E76A3",
            "name": "Balancer (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "BAL",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x03CD157746c61F44597dD54C6f6702105258C722",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x9c2C5fd7b07E95EE044DDeba0E97a665F142394f",
            "name": "1Inch (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "1INCH",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x443C5116CdF663Eb387e72C688D276e702135C87",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xA041544fe2BE56CCe31Ebb69102B965E06aacE80",
            "name": "BarnBridge Governance Token (PoS)",
            "releases": [],
            "symbol": "BOND",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0xA1c57f48F0Deb89f569dFbE6E2B7f46D33606fD4",
            "name": "(PoS) Decentraland MANA",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "MANA",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xA1CbF3Fe43BC3501e3Fc4b573e822c70e76A7512",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xa9f37D84c856fDa3812ad0519Dad44FA0a3Fe207",
            "name": "Enzyme (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "MLN",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xB89D583B72aBF9C3a7e6e093251C2fCad3365312",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xb33EaAd8d922B1083446DC23f610c2567fB5180f",
            "name": "Uniswap (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "UNI",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x162d8c5bF15eB6BEe003a1ffc4049C92114bc931",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xBbba073C31bF03b8ACf7c28EF0738DeCF3695683",
            "name": "SAND",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "SAND",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x3D49406EDd4D52Fb7FFd25485f32E073b529C924",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0xc2132D05D31c914a87C6611C10748AEb04B58e8F",
            "name": "(PoS) Tether USD",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "USDT",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x0A6513e40db6EB1b165753AD52E80663aeA50545",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xc26D47d5c33aC71AC5CF9F776D63Ba292a4F7842",
            "name": "Bancor Network Token (PoS)",
            "releases": [],
            "symbol": "BNT",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0xc4195D4060DaEac44058Ed668AA5EfEc50D77ff6",
            "name": "Aave BAL",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "amBAL",
            "type": "aave-v2",
            "underlying": "0x9a71012B13CA4d3D0Cdc72A177DF3ef03b0E76A3",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x03CD157746c61F44597dD54C6f6702105258C722",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0x8ffDf2DE812095b1D19CB146E4c004587C0A0692",
            "name": "Aave Polygon BAL",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolBAL",
            "type": "aave-v3",
            "underlying": "0x9a71012B13CA4d3D0Cdc72A177DF3ef03b0E76A3",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x03CD157746c61F44597dD54C6f6702105258C722",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xD6DF932A45C0f255f85145f286eA0b292B21C90B",
            "name": "Aave (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "AAVE",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xbE23a3AA13038CfC28aFd0ECe4FdE379fE7fBfc4",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 4,
            "id": "0xD85d1e945766Fea5Eda9103F918Bd915FbCa63E6",
            "name": "Celsius (PoS)",
            "releases": [],
            "symbol": "CEL",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0xDA537104D6A5edd53c6fBba9A898708E465260b6",
            "name": "(PoS) yearn.finance",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "YFI",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x9896A1eA7A00F5f32Ab131eBbeE07487B0af31D0",
                "rate_asset": RateAsset.ETH,
            },
        },
        {
            "decimals": 18,
            "id": "0xdAb529f40E671A1D4bF91361c21bf9f0C9712ab7",
            "name": "(PoS) Binance USD",
            "releases": [],
            "symbol": "BUSD",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0xe1708AbDE4847B4929b70547E5197F1Ba1db2250",
            "name": "Tokemak (PoS)",
            "releases": [],
            "symbol": "TOKE",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0xe4Bf2864ebeC7B7fDf6Eeca9BaCAe7cDfDAffe78",
            "name": "DODO bird (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "DODO",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x59161117086a4C7A9beDA16C66e40Bdaa1C5a8B6",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x5FFD62D3C3eE2E81C00A7b9079FB248e7dF024A8",
            "name": "Gnosis Token (PoS)",
            "releases": [],
            "symbol": "GNO",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x553d3D295e0f695B9228246232eDF400ed3560B5",
            "name": "Paxos Gold (PoS)",
            "releases": [POLYGON["sulu"]],
            "symbol": "PAXG",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x0f6914d8e7e1214CDb3A4C6fbf729b75C69DF608",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 6,
            "id": "0x7BDF330f423Ea880FF95fC41A280fD5eCFD3D09f",
            "name": "Euro Tether (PoS)",
            "releases": [],
            "symbol": "EURT",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x2e1AD108fF1D8C782fcBbB89AAd783aC49586756",
            "name": "TrueUSD (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "TUSD",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x7C5D415B64312D38c56B54358449d0a4058339d2",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x6F3B3286fd86d8b47EC737CEB3D0D354cc657B3e",
            "name": "Paxos Standard (PoS)",
            "releases": [],
            "symbol": "PAX",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0xE7a24EF0C5e95Ffb0f6684b813A78F2a3AD7D171",
            "name": "Curve.fi amDAI/amUSDC/amUSDT",
            "pool": "0x445FE580eF8d70FF569aB36e80c647af338db351",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "staking": [
                {
                    "token": "0x20759F567BB3EcDB55c817c9a1d13076aB215EdC",
                    type: "gauge",
                }
            ],
            "symbol": "am3CRV",
            "template": "aave",
            "type": "curve-pool-lp",
            "underlyings": [
                "0x27F8D03b3a2196956ED754baDc28D73be8830A6e",
                "0x1a13F4Ca1d028320A707D99520AbFefca3998b7F",
                "0x60D55F02A771d515e077c9C2403a1ef324885CeC",
            ],
            "price_feed": {
                "type": "DERIVATIVE_CURVE",
                "address": "0xc537439f4397a0f625Cb323F90BF87397dA92fCF",
                "ipa": "0x51e75b5E0eef2d40B4D70C5dAa2666E1eA30F0Bd",
            },
        },
        {
            "decimals": 18,
            "id": "0x20759F567BB3EcDB55c817c9a1d13076aB215EdC",
            "lp": "0xE7a24EF0C5e95Ffb0f6684b813A78F2a3AD7D171",
            "name": "Curve.fi am3CRV Gauge Deposit",
            "pool": "0x445FE580eF8d70FF569aB36e80c647af338db351",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "am3CRV-gauge",
            "template": "aave",
            "type": "curve-pool-gauge",
            "underlyings": [
                "0x27F8D03b3a2196956ED754baDc28D73be8830A6e",
                "0x1a13F4Ca1d028320A707D99520AbFefca3998b7F",
                "0x60D55F02A771d515e077c9C2403a1ef324885CeC",
            ],
            "price_feed": {
                "type": "DERIVATIVE_CURVE",
                "address": "0xc537439f4397a0f625Cb323F90BF87397dA92fCF",
                "ipa": "0x51e75b5E0eef2d40B4D70C5dAa2666E1eA30F0Bd",
            },
        },
        {
            "decimals": 18,
            "id": "0x5e5A23b52Cb48F5E70271Be83079cA5bC9c9e9ac",
            "name": "Curve.fi Factory USD Metapool: Curve FRAX3CRV Pool",
            "pool": "0x5e5A23b52Cb48F5E70271Be83079cA5bC9c9e9ac",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "staking": [
                {
                    "token": "0x78DA40dbFA5c3CC65319e8d658d3c9BaD3322f26",
                    type: "gauge",
                }
            ],
            "symbol": "FRAX3CRV-f3CRV-f",
            "template": "meta",
            "type": "curve-pool-lp",
            "underlyings": [
                "0x45c32fA6DF82ead1e2EF74d17b76547EDdFaFF89",
                "0xE7a24EF0C5e95Ffb0f6684b813A78F2a3AD7D171",
            ],
            "price_feed": {
                "type": "DERIVATIVE_CURVE",
                "address": "0xc537439f4397a0f625Cb323F90BF87397dA92fCF",
                "ipa": "0x51e75b5E0eef2d40B4D70C5dAa2666E1eA30F0Bd",
            },
        },
        {
            "decimals": 18,
            "id": "0x78DA40dbFA5c3CC65319e8d658d3c9BaD3322f26",
            "lp": "0x5e5A23b52Cb48F5E70271Be83079cA5bC9c9e9ac",
            "name": "Curve.fi FRAX3CRV-f3CRV-f Gauge Deposit",
            "pool": "0x5e5A23b52Cb48F5E70271Be83079cA5bC9c9e9ac",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "FRAX3CRV-f3CRV-f-gauge",
            "template": "meta",
            "type": "curve-pool-gauge",
            "underlyings": [
                "0x45c32fA6DF82ead1e2EF74d17b76547EDdFaFF89",
                "0xE7a24EF0C5e95Ffb0f6684b813A78F2a3AD7D171",
            ],
            "price_feed": {
                "type": "DERIVATIVE_CURVE",
                "address": "0xc537439f4397a0f625Cb323F90BF87397dA92fCF",
                "ipa": "0x51e75b5E0eef2d40B4D70C5dAa2666E1eA30F0Bd",
            },
        },
        {
            "decimals": 18,
            "id": "0x447646e84498552e62eCF097Cc305eaBFFF09308",
            "name": "Curve.fi Factory USD Metapool: MAI+3Pool",
            "pool": "0x447646e84498552e62eCF097Cc305eaBFFF09308",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "staking": [
                {
                    "token": "0x93DB264Ee9c42092DC963517885be167B5BF62DB",
                    type: "gauge",
                }
            ],
            "symbol": "MAI+3Pool3CRV-f",
            "template": "meta",
            "type": "curve-pool-lp",
            "underlyings": [
                "0xa3Fa99A148fA48D14Ed51d610c367C61876997F1",
                "0xE7a24EF0C5e95Ffb0f6684b813A78F2a3AD7D171",
            ],
            "price_feed": {
                "type": "DERIVATIVE_CURVE",
                "address": "0xc537439f4397a0f625Cb323F90BF87397dA92fCF",
                "ipa": "0x51e75b5E0eef2d40B4D70C5dAa2666E1eA30F0Bd",
            },
        },
        {
            "decimals": 18,
            "id": "0x93DB264Ee9c42092DC963517885be167B5BF62DB",
            "lp": "0x447646e84498552e62eCF097Cc305eaBFFF09308",
            "name": "Curve.fi MAI+3Pool3CRV-f Gauge Deposit",
            "pool": "0x447646e84498552e62eCF097Cc305eaBFFF09308",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "MAI+3Pool3CRV-f-gauge",
            "template": "meta",
            "type": "curve-pool-gauge",
            "underlyings": [
                "0xa3Fa99A148fA48D14Ed51d610c367C61876997F1",
                "0xE7a24EF0C5e95Ffb0f6684b813A78F2a3AD7D171",
            ],
            "price_feed": {
                "type": "DERIVATIVE_CURVE",
                "address": "0xc537439f4397a0f625Cb323F90BF87397dA92fCF",
                "ipa": "0x51e75b5E0eef2d40B4D70C5dAa2666E1eA30F0Bd",
            },
        },
        {
            "decimals": 18,
            "id": "0xAdf577B69eEaC9Df325536cf1aF106372f2Da263",
            "name": "Curve.fi Factory USD Metapool: curve.fi TUSD/am3CRV Pool",
            "pool": "0xAdf577B69eEaC9Df325536cf1aF106372f2Da263",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "staking": [
                {
                    "token": "0xdA690c2EA49a058a9966C69f46a05Bfc225939f4",
                    type: "gauge",
                }
            ],
            "symbol": "TUSDam3CRV3CRV-f",
            "template": "meta",
            "type": "curve-pool-lp",
            "underlyings": [
                "0x2e1AD108fF1D8C782fcBbB89AAd783aC49586756",
                "0xE7a24EF0C5e95Ffb0f6684b813A78F2a3AD7D171",
            ],
            "price_feed": {
                "type": "DERIVATIVE_CURVE",
                "address": "0xc537439f4397a0f625Cb323F90BF87397dA92fCF",
                "ipa": "0x51e75b5E0eef2d40B4D70C5dAa2666E1eA30F0Bd",
            },
        },
        {
            "decimals": 18,
            "id": "0xdA690c2EA49a058a9966C69f46a05Bfc225939f4",
            "lp": "0xAdf577B69eEaC9Df325536cf1aF106372f2Da263",
            "name": "Curve.fi TUSDam3CRV3CRV-f Gauge Deposit",
            "pool": "0xAdf577B69eEaC9Df325536cf1aF106372f2Da263",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "TUSDam3CRV3CRV-f-gauge",
            "template": "meta",
            "type": "curve-pool-gauge",
            "underlyings": [
                "0x2e1AD108fF1D8C782fcBbB89AAd783aC49586756",
                "0xE7a24EF0C5e95Ffb0f6684b813A78F2a3AD7D171",
            ],
            "price_feed": {
                "type": "DERIVATIVE_CURVE",
                "address": "0xc537439f4397a0f625Cb323F90BF87397dA92fCF",
                "ipa": "0x51e75b5E0eef2d40B4D70C5dAa2666E1eA30F0Bd",
            },
        },
        {
            "decimals": 18,
            "id": "0xa3Fa99A148fA48D14Ed51d610c367C61876997F1",
            "name": "miMATIC",
            "releases": [],
            "symbol": "miMATIC",
            "type": "primitive",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0xeBe517846d0F36eCEd99C735cbF6131e1fEB775D",
            "name": "Aave Polygon MIMATIC",
            "releases": [],
            "symbol": "aPolMIMATIC",
            "type": "aave-v3",
            "underlying": "0xa3Fa99A148fA48D14Ed51d610c367C61876997F1",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x8a999F5A3546F8243205b2c0eCb0627cC10003ab",
            "name": "IdleDAI [Best yield]",
            "releases": [],
            "symbol": "idleDAIYield",
            "type": "idle",
            "underlying": "0x8f3Cf7ad23Cd3CaDbD9735AFf958023239c6A063",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0xfa68FB4628DFF1028CFEc22b4162FCcd0d45efb6",
            "name": "Liquid Staking Matic (PoS)",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "MaticX",
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x5d37E4b374E6907de8Fc7fb33EE3b0af403C7403",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0x80cA0d8C38d2e2BcbaB66aA1648Bd1C7160500FE",
            "name": "Aave Polygon MATICX",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "aPolMATICX",
            "type": "aave-v3",
            "underlying": "0xfa68FB4628DFF1028CFEc22b4162FCcd0d45efb6",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x5d37E4b374E6907de8Fc7fb33EE3b0af403C7403",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "decimals": 18,
            "id": "0xfdA25D931258Df948ffecb66b5518299Df6527C4",
            "name": "IdleWETH [Best yield]",
            "releases": [],
            "symbol": "idleWETHYield",
            "type": "idle",
            "underlying": "0x7ceB23fD6bC0adD59E62ac25578270cFf1b9f619",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 18,
            "id": "0x1ee6470CD75D5686d0b2b90C0305Fa46fb0C89A1",
            "name": "IdleUSDC [Best yield]",
            "releases": [],
            "symbol": "idleUSDCYield",
            "type": "idle",
            "underlying": "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "decimals": 6,
            "id": "0xF25212E676D1F7F89Cd72fFEe66158f541246445",
            "name": "Compound USDC",
            "releases": [POLYGON["sulu"], TESTNET["sulu"]],
            "symbol": "cUSDCv3",
            "type": "compound-v3",
            "underlying": "0x2791Bca1f2de4661ED88A30C99A7a9449Aa84174",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0xfE4A8cc5b5B2366C1B58Bea3858e81843581b2F7",
                "rate_asset": RateAsset.USD,
            },
        },
        {
            "id": "0x4954AbbaA3637f19C7594aa248bF5A2b5f98F093",
            "name": "Balancer DUSD-USDC-BPT Gauge Deposit",
            "releases": [TESTNET["sulu"]],
            "symbol": "DUSD-USDC-BPT-gauge",
            "type": "primitive",
            "decimals": 18,
            "price_feed": {
                "type": "NONE",
            },
        },
        {
            "id": "0x6a8742D7F417431AE4218e94678D91f2aCE24a66",
            "name": "UI Space Y Stock",
            "symbol": "UISTEST",
            "decimals": 0,
            "releases": [POLYGON["sulu"]],
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x73366Fe0AA0Ded304479862808e02506FE556a98",
                "rate_asset": RateAsset.USD,
                "pegged_to": "EUR",
                "non_standard": True,
            },
        },
        {
            "id": "0xDD52Fb974E8b486Da82e69d56DA8C6f8Bf35B0e0",
            "name": "UI Real Estate",
            "symbol": "UIRETEST",
            "decimals": 0,
            "releases": [POLYGON["sulu"]],
            "type": "primitive",
            "price_feed": {
                "type": "PRIMITIVE_CHAINLINK",
                "aggregator": "0x73366Fe0AA0Ded304479862808e02506FE556a98",
                "rate_asset": RateAsset.USD,
                "pegged_to": "EUR",
                "non_standard": True,
            },
        },
    ],
)



import asyncio
from typing import TypedDict, Any
from web3.types import ChecksumAddress, HexStr, TxParams
from .utils.clients import PublicClient, WalletClient
from . import asset as assets
from .portfolio_lib import asset_managers
from .portfolio_lib import integrations
from ..abis import abis

from ._internal.external_position_manager import (
    ACTION as EXTERNAL_POSITION_ACTION,
    call as call_external_position,
    call_encode as call_external_position_encode,
    call_decode as call_external_position_decode,
    CallArgs as ExternalPositionCallArgs,
    create as create_external_position,
    create_encode as create_external_position_encode,
    create_decode as create_external_position_decode,
    CreateArgs as CreateExternalPositionArgs,
    remove as remove_external_position,
    remove_encode as remove_external_position_encode,
    remove_decode as remove_external_position_decode,
    RemoveArgs as RemoveExternalPositionArgs,
    reactivate as reactivate_external_position,
    reactivate_encode as reactivate_external_position_encode,
    reactivate_decode as reactivate_external_position_decode,
    ReactivateArgs as ReactivateExternalPositionArgs,
)


from ._internal.integration_manager import (
    ACTION as INTEGRATION_ADAPTER_ACTION,
    SELECTOR as INTEGRATION_ADAPTER_SELECTOR,
    call as call_integration_adapter,
    call_encode as call_integration_adapter_encode,
    call_decode as call_integration_adapter_decode,
    CallArgs as CallIntegrationAdapterArgs,
    add_tracked_assets,
    add_tracked_assets_encode,
    add_tracked_assets_decode,
    AddTrackedAssetsArgs,
    remove_tracked_assets,
    remove_tracked_assets_encode,
    remove_tracked_assets_decode,
    RemoveTrackedAssetsArgs,
)


class VaultCallOnContractParams(TypedDict):
    client: WalletClient
    comptroller_proxy: ChecksumAddress
    contract: ChecksumAddress
    selector: HexStr
    encoded_args: HexStr


async def vault_call_on_contract(
    client: WalletClient,
    comptroller_proxy: ChecksumAddress,
    contract: ChecksumAddress,
    selector: HexStr,
    encoded_args: HexStr,
) -> TxParams:
    contract = client.contract(comptroller_proxy, abis.IComptrollerLib)
    function = contract.functions.vaultCallOnContract(contract, selector, encoded_args)
    return await client.populated_transaction(function)


async def get_portfolio(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> dict[str, Any]:
    external_position_addresses, tracked_asset_addresses = await asyncio.gather(
        get_active_external_positions(client, vault_proxy),
        get_tracked_assets(client, vault_proxy),
    )

    external_positions_data, tracked_assets_amounts = await asyncio.gather(
        asyncio.gather(
            *[
                get_external_position_type(client, pos)
                for pos in external_position_addresses
            ],
            *[
                get_external_position_debt_assets(client, pos)
                for pos in external_position_addresses
            ],
            *[
                get_external_position_managed_assets(client, pos)
                for pos in external_position_addresses
            ],
        ),
        asyncio.gather(
            *[
                assets.get_balance_of(client, vault_proxy, asset)
                for asset in tracked_asset_addresses
            ],
        ),
    )

    pos_count = len(external_position_addresses)
    types = external_positions_data[0:pos_count]
    debts = external_positions_data[pos_count : pos_count * 2]
    managed = external_positions_data[pos_count * 2 :]

    external_positions_data = [
        {
            "external_position": pos,
            "external_position_type": typ,
            "debt_assets": debt,
            "managed_assets": mng,
        }
        for pos, typ, debt, mng in zip(
            external_position_addresses, types, debts, managed
        )
    ]

    tracked_assets_data = [
        {"asset": asset, "amount": balance}
        for asset, balance in zip(tracked_asset_addresses, tracked_assets_amounts)
    ]

    return {
        "external_positions": external_positions_data,
        "tracked_assets": tracked_assets_data,
    }


async def get_tracked_assets(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> list[ChecksumAddress]:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.getTrackedAssets()
    return await function.call()


# --------------------------------------------------------------------------------------------
# EXTERNAL POSITIONS
# --------------------------------------------------------------------------------------------


async def is_active_external_position(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
    external_position: ChecksumAddress,
) -> bool:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.isActiveExternalPosition(external_position)
    return await function.call()


async def get_total_value_for_all_external_positions(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> int:
    addresses = await get_active_external_positions(client, vault_proxy)
    external_position_assets = await asyncio.gather(
        *[get_external_position_assets(client, addr) for addr in addresses]
    )
    values = []
    for assets in external_position_assets:
        debt_assets = assets["debt_assets"]
        managed_assets = assets["managed_assets"]

        debt_assets_value = sum(asset["amount"] for asset in debt_assets)
        managed_assets_values = sum(asset["amount"] for asset in managed_assets)

        if managed_assets_values > debt_assets_value:
            values.append(managed_assets_values - debt_assets_value)
        else:
            values.append(0)

    return sum(values)


async def get_active_external_positions(
    client: PublicClient,
    vault_proxy: ChecksumAddress,
) -> list[ChecksumAddress]:
    contract = client.contract(vault_proxy, abis.IVaultLib)
    function = contract.functions.getActiveExternalPositions()
    return await function.call()


async def get_external_position_managed_assets(
    client: PublicClient,
    external_position: ChecksumAddress,
) -> list[dict[ChecksumAddress, int]]:
    contract = client.contract(external_position, abis.IExternalPosition)
    function = contract.functions.getManagedAssets()
    assets, amounts = await function.call()
    assert all(amount is not None for amount in amounts), "Missing managed asset amount"
    return [
        {
            "asset": asset,
            "amount": amount,
        }
        for asset, amount in zip(assets, amounts)
    ]


async def get_external_position_debt_assets(
    client: PublicClient,
    external_position: ChecksumAddress,
) -> list[dict[ChecksumAddress, int]]:
    contract = client.contract(external_position, abis.IExternalPosition)
    function = contract.functions.getDebtAssets()
    assets, amounts = await function.call()
    assert all(amount is not None for amount in amounts), "Missing debt asset amount"
    return [
        {
            "asset": asset,
            "amount": amount,
        }
        for asset, amount in zip(assets, amounts)
    ]


async def get_external_position_assets(
    client: PublicClient,
    external_position: ChecksumAddress,
) -> dict[str, list[dict[ChecksumAddress, int]]]:
    debt_assets, managed_assets = await asyncio.gather(
        get_external_position_debt_assets(client, external_position),
        get_external_position_managed_assets(client, external_position),
    )
    return {
        "debt_assets": debt_assets,
        "managed_assets": managed_assets,
    }


async def get_external_position_type(
    client: PublicClient,
    external_position: ChecksumAddress,
) -> int:
    contract = client.contract(external_position, abis.IExternalPositionProxy)
    function = contract.functions.getExternalPositionType()
    return await function.call()


async def get_type_label(
    client: PublicClient,
    external_position_factory: ChecksumAddress,
    type_id: int,
) -> str:
    contract = client.contract(external_position_factory, abis.IExternalPositionFactory)
    function = contract.functions.getTypeLabel(type_id)
    return await function.call()



from typing import Callable, Any
from web3.constants import ADDRESS_ZERO
from web3.types import ChecksumAddress
from .adapters import get_adapters_for_release
from .assets import Asset, AssetType
from .contracts import is_version, Version, VersionContractNames
from .networks import get_network
from .releases import DeploymentDefinition, Version, Deployment


class Environment:
    def __init__(self, deployment: DeploymentDefinition, version: Version):
        self.deployment = deployment
        self.version = version

        release = deployment["releases"][version]
        if not release:
            raise ValueError(
                f"Invalid release {version} for {deployment['slug']} deployment"
            )

        self.external_contracts = deployment["external_contracts"]
        self.known_address_lists = deployment["known_address_lists"]
        self.known_uint_lists = deployment["known_uint_lists"]

        network = deployment["network"]
        assets = [
            asset for asset in deployment["assets"] if asset["network"] == network
        ]

        self.release = release
        self.contracts = release["contracts"]
        self.network = get_network(network)
        self.assets: dict[ChecksumAddress, Asset] = {
            asset["id"]: asset
            | {"network": network, "registered": release["slug"] in asset["releases"]}
            for asset in assets
        }

        self.named_tokens = self._initialize_named_tokens()
        self.adapters = get_adapters_for_release(self.contracts)

    def _initialize_named_tokens(self) -> dict[str, Asset]:
        if self.is_deployment_arbitrum(self):
            return {
                "bal": self.get_asset_as(
                    self.deployment["named_tokens"]["bal"], "primitive"
                ),
                "comp": self.get_asset_as(
                    self.deployment["named_tokens"]["comp"], "primitive"
                ),
                "crv": self.get_asset_as(
                    self.deployment["named_tokens"]["crv"], "primitive"
                ),
                "cvx": self.get_asset_as(
                    self.deployment["named_tokens"]["cvx"], "primitive"
                ),
                "dai": self.get_asset_as(
                    self.deployment["named_tokens"]["dai"], "primitive"
                ),
                "grt": self.get_asset_as(
                    self.deployment["named_tokens"]["grt"], "primitive"
                ),
                "mln": self.get_asset_as(
                    self.deployment["named_tokens"]["mln"], "primitive"
                ),
                "native_token_wrapper": self.get_asset_as(
                    self.network["currency"]["wrapper"], "primitive"
                ),
                "usdt": self.get_asset_as(
                    self.deployment["named_tokens"]["usdt"], "primitive"
                ),
                "weth": self.get_asset_as(
                    self.deployment["named_tokens"]["weth"], "primitive"
                ),
            }
        elif self.is_deployment_base(self):
            return {
                "comp": self.get_asset_as(
                    self.deployment["named_tokens"]["comp"], "primitive"
                ),
                "dai": self.get_asset_as(
                    self.deployment["named_tokens"]["dai"], "primitive"
                ),
                "mln": self.get_asset_as(
                    self.deployment["named_tokens"]["mln"], "primitive"
                ),
                "native_token_wrapper": self.get_asset_as(
                    self.network["currency"]["wrapper"], "primitive"
                ),
                "usdt": self.get_asset_as(
                    self.deployment["named_tokens"]["usdt"], "primitive"
                ),
                "weth": self.get_asset_as(
                    self.deployment["named_tokens"]["weth"], "primitive"
                ),
            }
        elif self.is_deployment_ethereum(self):
            return {
                "aave": self.get_asset_as(
                    self.deployment["named_tokens"]["aave"], "primitive"
                ),
                "bal": self.get_asset_as(
                    self.deployment["named_tokens"]["bal"], "primitive"
                ),
                "ceth": self.get_asset_as(
                    self.deployment["named_tokens"]["ceth"], "compound_v2"
                ),
                "comp": self.get_asset_as(
                    self.deployment["named_tokens"]["comp"], "primitive"
                ),
                "crv": self.get_asset_as(
                    self.deployment["named_tokens"]["crv"], "primitive"
                ),
                "cvx": self.get_asset_as(
                    self.deployment["named_tokens"]["cvx"], "primitive"
                ),
                "dai": self.get_asset_as(
                    self.deployment["named_tokens"]["dai"], "primitive"
                ),
                "diva": self.get_asset_as(
                    self.deployment["named_tokens"]["diva"], "primitive"
                ),
                "ethx": self.get_asset_as(
                    self.deployment["named_tokens"]["ethx"], "primitive"
                ),
                "grt": self.get_asset_as(
                    self.deployment["named_tokens"]["grt"], "primitive"
                ),
                "idle": self.get_asset_as(
                    self.deployment["named_tokens"]["idle"], "primitive"
                ),
                "lusd": self.get_asset_as(
                    self.deployment["named_tokens"]["lusd"], "primitive"
                ),
                "mln": self.get_asset_as(
                    self.deployment["named_tokens"]["mln"], "primitive"
                ),
                "mpl": self.get_asset_as(
                    self.deployment["named_tokens"]["mpl"], "primitive"
                ),
                "native_token_wrapper": self.get_asset_as(
                    self.network["currency"]["wrapper"], "primitive"
                ),
                "paxg": self.get_asset_as(
                    self.deployment["named_tokens"]["paxg"], "primitive"
                ),
                "ptkn_mln": self.get_asset_as(
                    self.deployment["named_tokens"]["ptkn_mln"], "primitive"
                ),
                "sthoundeth": self.get_asset_as(
                    self.deployment["named_tokens"]["sthoundeth"], "primitive"
                ),
                "stkaave": self.get_asset_as(
                    self.deployment["named_tokens"]["stkaave"], "primitive"
                ),
                "steth": self.get_asset_as(
                    self.deployment["named_tokens"]["steth"], "primitive"
                ),
                "stusd": self.get_asset_as(
                    self.deployment["named_tokens"]["stusd"], "erc_4626"
                ),
                "sweth": self.get_asset_as(
                    self.deployment["named_tokens"]["sweth"], "primitive"
                ),
                "uni": self.get_asset_as(
                    self.deployment["named_tokens"]["uni"], "primitive"
                ),
                "usda": self.get_asset_as(
                    self.deployment["named_tokens"]["usda"], "primitive"
                ),
                "usdc": self.get_asset_as(
                    self.deployment["named_tokens"]["usdc"], "primitive"
                ),
                "usdt": self.get_asset_as(
                    self.deployment["named_tokens"]["usdt"], "primitive"
                ),
                "weth": self.get_asset_as(
                    self.deployment["named_tokens"]["weth"], "primitive"
                ),
            }
        elif self.is_deployment_polygon(self) or self.is_deployment_testnet(self):
            return {
                "aave": self.get_asset_as(
                    self.deployment["named_tokens"]["aave"], "primitive"
                ),
                "bal": self.get_asset_as(
                    self.deployment["named_tokens"]["bal"], "primitive"
                ),
                "comp": self.get_asset_as(
                    self.deployment["named_tokens"]["comp"], "primitive"
                ),
                "crv": self.get_asset_as(
                    self.deployment["named_tokens"]["crv"], "primitive"
                ),
                "cvx": self.get_asset_as(
                    self.deployment["named_tokens"]["cvx"], "primitive"
                ),
                "dai": self.get_asset_as(
                    self.deployment["named_tokens"]["dai"], "primitive"
                ),
                "eure": self.get_asset_as(
                    self.deployment["named_tokens"]["eure"], "primitive"
                ),
                "grt": self.get_asset_as(
                    self.deployment["named_tokens"]["grt"], "primitive"
                ),
                "mln": self.get_asset_as(
                    self.deployment["named_tokens"]["mln"], "primitive"
                ),
                "native_token_wrapper": self.get_asset_as(
                    self.network["currency"]["wrapper"], "primitive"
                ),
                "uni": self.get_asset_as(
                    self.deployment["named_tokens"]["uni"], "primitive"
                ),
                "usdc": self.get_asset_as(
                    self.deployment["named_tokens"]["usdc"], "primitive"
                ),
                "usdt": self.get_asset_as(
                    self.deployment["named_tokens"]["usdt"], "primitive"
                ),
                "weth": self.get_asset_as(
                    self.deployment["named_tokens"]["weth"], "primitive"
                ),
            }
        else:
            raise ValueError("Invalid deployment")

    @staticmethod
    def _create_is_version(version: Version) -> Callable[["Environment"], bool]:
        return lambda env: env.release["version"] == version

    is_sulu = staticmethod(_create_is_version("sulu"))
    is_encore = staticmethod(_create_is_version("encore"))
    is_phoenix = staticmethod(_create_is_version("phoenix"))

    @staticmethod
    def is_version(version: Version, environment: "Environment") -> bool:
        return environment.release["version"] == version

    @staticmethod
    def _create_is_deployment(
        deployment: Deployment,
    ) -> Callable[["Environment"], bool]:
        return lambda env: env.deployment["slug"] == deployment

    is_deployment_arbitrum = staticmethod(_create_is_deployment("arbitrum"))
    is_deployment_base = staticmethod(_create_is_deployment("base"))
    is_deployment_ethereum = staticmethod(_create_is_deployment("ethereum"))
    is_deployment_polygon = staticmethod(_create_is_deployment("polygon"))
    is_deployment_testnet = staticmethod(_create_is_deployment("testnet"))

    @staticmethod
    def is_deployment(deployment: Deployment, environment: "Environment") -> bool:
        return environment.deployment["slug"] == deployment

    def has_asset(self, address: ChecksumAddress) -> bool:
        if address == ADDRESS_ZERO:
            raise ValueError(f"Invalid address {address}")
        return address in self.assets

    def get_asset(self, address: ChecksumAddress) -> Asset:
        if not self.has_asset(address):
            raise ValueError(f"Invalid asset {address}")
        return self.assets[address]

    def get_asset_as(self, address: ChecksumAddress, type_: AssetType) -> Asset:
        if not self.has_asset(address):
            raise ValueError(f"Invalid asset {address}")
        asset = self.assets[address]
        if asset["type"] != type_:
            raise ValueError(
                f"Invalid asset type: Expected {type_} but got {asset['type']}"
            )
        return asset

    def get_assets(
        self, filter_: dict[str, bool | list[AssetType]] | None = None
    ) -> list[Asset]:
        filter_ = filter_ or {}
        types = filter_.get("types", [])
        registered = filter_.get("registered")

        assets = list(self.assets.values())

        if types:
            assets = [a for a in assets if a["type"] in types] if types else []

        if registered is not None:
            assets = [a for a in assets if a["registered"] == registered]

        return sorted(assets, key=lambda x: x["name"])

    def has_contract(self, name: VersionContractNames) -> bool:
        return self.contracts.get(name, ADDRESS_ZERO) != ADDRESS_ZERO

    def get_contract(self, name: VersionContractNames) -> ChecksumAddress:
        if not self.has_contract(name):
            raise ValueError(f"Missing contract {name}")
        return self.contracts[name]

    def __str__(self) -> str:
        return f"{self.deployment['slug']}.{self.version}"

    def to_json(self) -> str:
        return self.__str__()


class EnvironmentGroup:
    def __init__(self, deployment: DeploymentDefinition):
        self.deployment = deployment
        self.network = get_network(deployment["network"])
        self.assets: dict[ChecksumAddress, Asset] = {
            asset["id"]: asset
            for asset in deployment["assets"]
            if asset["network"] == self.network["id"]
        }
        self._environments: dict[Version, Environment] = {}

    def has_environment(self, version_or_address: ChecksumAddress | Version) -> bool:
        if is_version(version_or_address):
            return version_or_address in self.deployment["releases"]
        return any(
            release["address"] == version_or_address
            for release in self.deployment["releases"].values()
        )

    def get_version(self, address: ChecksumAddress) -> Version:
        for release in self.deployment["releases"].values():
            if release["address"] == address:
                return release["version"]
        raise ValueError(f"No release found for address {address}")

    def get_environment(
        self, version_or_address: ChecksumAddress | Version
    ) -> Environment:
        version = (
            version_or_address
            if is_version(version_or_address)
            else self.get_version(version_or_address)
        )

        if version not in self._environments and version in self.deployment["releases"]:
            self._environments[version] = Environment(self.deployment, version)

        if version not in self._environments:
            raise ValueError(
                f"Invalid release {version} for deployment {self.deployment['slug']}"
            )

        return self._environments[version]  # type: ignore

    @property
    def sulu(self) -> Environment:
        return self.get_environment("sulu")

    @property
    def encore(self) -> Environment:
        return self.get_environment("encore")

    @property
    def phoenix(self) -> Environment:
        return self.get_environment("phoenix")



from web3 import Web3
from web3.types import ChecksumAddress, HexStr, TxParams
from eth_abi import encode, decode
from ..._internal import integration_manager as integration_manager_lib
from ...utils.encoding import encoding_to_types
from typing import Any, TypedDict

# --------------------------------------------------------------------------------------------
# TAKE ORDER
# --------------------------------------------------------------------------------------------


async def take_order(*args) -> TxParams:
    """
    Args:
        client: WalletClient
        comptroller_proxy: ChecksumAddress
        integration_manager: ChecksumAddress
        integration_adapter: ChecksumAddress
        call_args: dict[str, Any]
            {
                "path_addresses": list[ChecksumAddress],
                "path_fees": list[int],
                "outgoing_asset_amount": int,
                "min_incoming_asset_amount": int,
            }
    """
    _take_order = integration_manager_lib.make_use(
        integration_manager_lib.SELECTOR["take_order"],
        take_order_encode,
    )
    return await _take_order(*args)


TAKE_ORDER_ENCODING = [
    {
        "name": "pathAddresses",
        "type": "address[]",
    },
    {
        "name": "pathFees",
        "type": "uint24[]",
    },
    {
        "name": "outgoingAssetAmount",
        "type": "uint256",
    },
    {
        "name": "minIncomingAssetAmount",
        "type": "uint256",
    },
]


class TakeOrderArgs(TypedDict):
    path_addresses: list[ChecksumAddress]
    path_fees: list[int]
    outgoing_asset_amount: int
    min_incoming_asset_amount: int


def take_order_encode(
    path_addresses: list[ChecksumAddress],
    path_fees: list[int],
    outgoing_asset_amount: int,
    min_incoming_asset_amount: int,
) -> HexStr:
    types = encoding_to_types(TAKE_ORDER_ENCODING)
    values = [
        path_addresses,
        path_fees,
        outgoing_asset_amount,
        min_incoming_asset_amount,
    ]
    return Web3.to_hex(encode(types, values))


def take_order_decode(encoded: HexStr) -> dict[str, Any]:
    """
    Returns:
        dict[str, Any]
            {
                "path_addresses": list[ChecksumAddress],
                "path_fees": list[int],
                "outgoing_asset_amount": int,
                "min_incoming_asset_amount": int,
            }
    """
    types = encoding_to_types(TAKE_ORDER_ENCODING)
    decoded = decode(types, Web3.to_bytes(hexstr=encoded))
    return {
        "path_addresses": [Web3.to_checksum_address(address) for address in decoded[0]],
        "path_fees": decoded[1],
        "outgoing_asset_amount": decoded[2],
        "min_incoming_asset_amount": decoded[3],
    }
